#include <Arduino.h>
#include <Wire.h>
#include "EEPROM24FC.h"

// コレを関数化しないと流石にメンドクサイので切り出しておきます
// I2Cバスが使用可能になるのを最大500m秒待ちます。
// 500m秒待って使用可能にならなかった時のことは考えてません。
void EEPROM24FC::writeReadyCheck(uint8_t i2cAddr) {
  int errCount = 0 ;
  while(errCount < 100) {
    Wire.beginTransmission(i2cAddr);    // I2Cスレーブに対して送信処理を開始する
    int ready = Wire.endTransmission(); // トランザクションが正常に終了する事を持ってI2Cバスが使用可能と判断
    if (ready == 0) {
      break ;
    }
    if (errCount > 100) break ;
    errCount ++ ;
    delay(5) ;
    break;
  }
}

// 1Byte書込み
void EEPROM24FC::writeByteData(uint8_t i2cAddr , uint16_t addredss , uint8_t data) {
  uint8_t adrH = (addredss & 0xFF00) >> 8 ;
  uint8_t adrL = addredss & 0xFF ;
    
  // 書き込み
  writeReadyCheck(i2cAddr) ;
  // 書き込み開始アドレス送信
  Wire.beginTransmission(i2cAddr);    // I2Cスレーブに対して送信処理を開始する
  Wire.write(adrH);
  Wire.write(adrL);
  while (Wire.write(data) == 0) {
    // 書き込み失敗時のリトライ
    Wire.endTransmission();
    writeReadyCheck(i2cAddr) ;

    // 書き込み開始アドレス送信
    Wire.beginTransmission(i2cAddr);    // I2Cスレーブに対して送信処理を開始する
    Wire.write(adrH);
    Wire.write(adrL);
  }
  Wire.endTransmission();
}

// マルチバイト書き込み
void EEPROM24FC::writeMultiData(uint8_t i2cAddr ,uint16_t address , int dataSize , uint8_t data[]) {
  int wBytes = dataSize ;           // 64バイト書き込み
  int dataPos = 0 ;
 
  // 書き込み
  writeReadyCheck(i2cAddr) ;
 
  while (wBytes > 0) {
    // 書き込み開始アドレスを上位バイトと下位バイトに分解
    uint8_t adrH = (address & 0xFF00) >> 8 ;
    uint8_t adrL = address & 0xFF ;
    // 書き込み開始アドレス送信
    Wire.beginTransmission(i2cAddr);    // I2Cスレーブに対して送信処理を開始する
    Wire.write(adrH);
    Wire.write(adrL);
 
    // データ送信
    while (wBytes > 0) {
      uint8_t wdata = data[dataPos];    // 書き込むデータ
      if (Wire.write(wdata) == 0) {
        // Wireクラスの制限で連続して書き込めるのは32バイト、33バイト目の書き込みは必ず失敗するので
        // 失敗したらトランザクションを一旦閉じて、再度、書き込みアドレスの指定から始める
        Wire.endTransmission();

        // 書き込み
        writeReadyCheck(i2cAddr) ;

        // 書き込み開始アドレスを上位バイトと下位バイトに分解
        adrH = (address & 0xFF00) >> 8 ;
        adrL = address & 0xFF ;

        // 書き込み開始アドレス送信
        Wire.beginTransmission(i2cAddr);    // I2Cスレーブに対して送信処理を開始する
        Wire.write(adrH);
        Wire.write(adrL);
      }
      dataPos ++ ;
      wBytes -- ;
      address ++ ;
 
      if ((address & B00011111) == 0) {
        // ページ境界を超えられないので、ページ境界に達したら
        // トランザクションを一旦閉じて、再度、書き込みアドレスの指定から始める
        // EEPROM24FCのページサイズは128Byteのようですが
        // どのみちArduinoでは32Byteずつしか書き込みないのだから
        // 直さなくてもよいか～と。実行時間もさほど変わらないですし、バグる方が怖いのでこのまま
        Wire.endTransmission();

        // 書き込み
        adrH = (address & 0xFF00) >> 8 ;
        adrL = address & 0xFF ;
        writeReadyCheck(i2cAddr) ;

        // 書き込み開始アドレス送信
        Wire.beginTransmission(i2cAddr);    // I2Cスレーブに対して送信処理を開始する
        Wire.write(adrH);
        Wire.write(adrL);
      }
    }
  }
  Wire.endTransmission();
}

// 1バイト読み出し
uint8_t EEPROM24FC::readByteData(uint8_t i2cAddr ,uint16_t address ) {
  // 書き込み開始アドレスを上位バイトと下位バイトに分解
  uint8_t adrH = (address & 0xFF00) >> 8 ;
  uint8_t adrL = address & 0xFF ;

  writeReadyCheck(i2cAddr) ;
  // 書き込み開始アドレス送信
  Wire.beginTransmission(i2cAddr);    // I2Cスレーブに対して送信処理を開始する
  Wire.write(adrH);
  Wire.write(adrL);
  Wire.endTransmission();

  int len = Wire.requestFrom(i2cAddr, 1);
  uint8_t data = 0 ;
  if (len == 1) {
    data = Wire.read();
  }
  return data ;
}

// マルチバイト読み出し
void EEPROM24FC::readMultiData(uint8_t i2cAddr , uint16_t address , int dataSize , uint8_t* data) {
  // 書き込み開始アドレスを上位バイトと下位バイトに分解
  uint8_t adrH = (address & 0xFF00) >> 8 ;
  uint8_t adrL = address & 0xFF ;
    
  // 書き込み
  writeReadyCheck(i2cAddr) ;

  // 書き込み開始アドレス送信
  Wire.beginTransmission(i2cAddr);    // I2Cスレーブに対して送信処理を開始する
  Wire.write(adrH);
  Wire.write(adrL);
  Wire.endTransmission();

  int rBytes = dataSize ; // 読み込みデータ長
  // 読み込み開始
  while(rBytes>0) {
    // Wireクラスの制限で一度に読み込めるのは32バイトまでなので
    // 読み込むバイト数-読み込んだバイト数がゼロになるまで繰り返す
    // 書き込み開始アドレスを上位バイトと下位バイトに分解
    uint8_t adrH = (address & 0xFF00) >> 8 ;
    uint8_t adrL = address & 0xFF ;

    int len = Wire.requestFrom(i2cAddr, rBytes);
    for (int j=0;j<len;j++) {
      *data = Wire.read();
      data ++ ;
    }
    rBytes -= len ;
    address += len ;
  }
}
