// 美咲フォントROM作成　01
// 美咲フォントをEEPROM 24FC (1MBit)に書き込む

#include <Arduino.h>
#include <Wire.h>
#include "MISAKI.h"
#include "EEPROM24FC.h"

EEPROM24FC eeprom24fc ;

char str[128] ;

// UTF-16 CODE LIST OFFESET TABLE Write
void writeIndexTable(bool flagWrite) {
  Serial.println("\n\n---　FONT INDEX TABLE ---") ;

  // --- Write ---
  unsigned int addr = 0 ;
  int tblSize = sizeof(fIndexTable) / sizeof(fIndexTable[0]) ;
  if (flagWrite) {
    Serial.println("--- Write ---") ;
    for (int i=0;i<tblSize ; i++ ) {
      uint16_t data = pgm_read_byte(fIndexTable + i);
      eeprom24fc.writeByteData(ROM1 , i , data) ;
  
      sprintf(str,"%4d,",data) ;
      Serial.print(str) ;
      if (((i+1) % 16) == 0) {
        Serial.println() ;
      }
    }
    Serial.println() ;
    Serial.println() ;
  }

  // --- verify ---
  int verifyError = 0 ;
  Serial.println("--- verify ---") ;
  for (int i=0;i<tblSize ; i++ ) {
    uint8_t romData = eeprom24fc.readByteData(ROM1,i) ;
    uint8_t data = pgm_read_byte(fIndexTable + i);

    if (romData != data) {
      verifyError ++ ;
      sprintf(str,"★%4d,",romData) ;
    } else {
      sprintf(str,"%4d,",romData) ;
    }
    Serial.print(str) ;
    if (((i+1) % 16) == 0) {
      Serial.println() ;         
    }
  }
  Serial.print("\n--- verify --- ") ;
  if (verifyError > 0) {
    Serial.println("ERROR") ;
  } else {
    Serial.println("OK") ;
  }
}

// 半角全角変換テーブル Write
void writeHFTable(bool flagWrite) {
  Serial.println("\n\n---　半角全角変換テーブル Write ---") ;

  // --- Write ---
  uint16_t addr = 53248 ;
  int tblSize = sizeof(hfTable) / sizeof(hfTable[0]) ;
  if (flagWrite) {
    Serial.println("--- Write ---") ;
    for (int i=0;i<tblSize ; i++ ) {
      uint8_t data = pgm_read_byte(hfTable + i);
      eeprom24fc.writeByteData(ROM1,addr,data) ;
      addr ++ ;
  
      sprintf(str,"%02X,",data) ;
      Serial.print(str) ;
      if (((i+1) % 16) == 0) {
        Serial.println() ;
      }
    }
    Serial.println() ;
    Serial.println() ;
  }

  // --- verify ---
  int verifyError = 0 ;
  addr = 53248 ;
  Serial.println("--- verify ---") ;
  for (int i=0;i<tblSize ; i++ ) {
    uint8_t romData = eeprom24fc.readByteData(ROM1,addr) ;
    uint16_t data = pgm_read_byte(hfTable + i);
    addr ++ ;
    
    if (romData != data) {
      verifyError ++ ;
      sprintf(str,"★%02X,",romData) ;
    } else {
      sprintf(str,"%02X,",romData) ;
    }
    Serial.print(str) ;
    if (((i+1) % 16) == 0) {
      Serial.println() ;         
    }
  }
  Serial.print("\n--- verify --- ") ;
  if (verifyError > 0) {
    Serial.println("ERROR") ;
  } else {
    Serial.println("OK") ;
  }
}

// FONTデータ Write
// FONTデータはUTF-16文字コードの昇順にソートされていることを前提としている
// 1バイト目が同じ文字は連続して書き込まないとアドレスがズレる
// ソートされていないと文字コードINDEXが作れないので注意
void writeFont( uint8_t *fontTable , int fontNum , bool flagWriteCodeTable) {
  Serial.println("\n\n--- FONTデータ Write ---") ;
  uint8_t fontData[12] ;
  uint8_t *wFontTable = fontTable ;

  if (flagWriteCodeTable) {
    Serial.println("--- Write ---") ;
    for (int i = 0 ; i < fontNum ; i++) {

      for (int j=0;j<11;j++) {
        fontData[j] = pgm_read_byte(wFontTable);
        wFontTable ++ ;
      }
  
      sprintf(str,"UTF16[%02X %02X] - " ,fontData[2],fontData[3]) ;
      Serial.print(str) ;

      // ---  FONT DATA ADDRESS 書き込み
      uint8_t indexTblNo = eeprom24fc.readByteData(ROM1 ,(uint16_t)fontData[2]) ;
      uint16_t indexAddr = 512 * (uint16_t)indexTblNo + (uint16_t)fontData[3] * 2  ;
      uint16_t fontAddr = fontData[0] << 8 | fontData[1] ;
      eeprom24fc.writeMultiData(ROM1,indexAddr,2,(uint8_t *)&fontAddr) ;
      sprintf(str,"FONT INDEX %04X = 512 x %02X + %02d x 2 - ",indexAddr,indexTblNo, fontData[3] ) ;
      Serial.print(str) ;
      sprintf(str,"FONT ADDR[%04X:%02X %02X] - " ,indexAddr , fontData[0],fontData[1]) ;
      Serial.print(str) ;
      for (int j=4;j<11;j++) {
        sprintf(str,"%02X,",fontData[j]) ;
        Serial.print(str) ;
      }
      Serial.println() ;

      // ---  FONT PATTERN 書き込み
      fontAddr = (fontData[0] << 8) | fontData[1] ;
      eeprom24fc.writeMultiData(ROM2 , fontAddr, 7 , (uint8_t *)&fontData[4]) ;

    }
    Serial.println() ;
  }

  Serial.println("--- verify ---") ;
  wFontTable = fontTable ;
  int verifyError = 0 ;
  uint8_t fontDataRom[12] ;
  uint16_t fontAddr ;
  for (int i = 0 ; i < fontNum ; i++) {
    for (int j=0;j<11;j++) {
      fontData[j] = pgm_read_byte(wFontTable);
      wFontTable ++ ;
      sprintf(str,"%02X,",fontData[j]) ;
      Serial.print(str) ;
    }
    Serial.print(" - ") ;
    
    // ---  INDEX TABLE ADDR
    uint8_t data = eeprom24fc.readByteData(ROM1 ,fontData[2]) ;
    uint16_t indexAddr = 512 * data + (uint16_t)fontData[3] * 2  ;
    eeprom24fc.readMultiData(ROM1,indexAddr,7,(uint8_t *)&fontAddr) ;

    // ---  FONT PATTERN
    eeprom24fc.readMultiData(ROM2, fontAddr,7,fontDataRom) ;
    sprintf(str,"INDEX:[%04X:%04X] - ",indexAddr,fontAddr) ;
    Serial.print(str) ;
    for (int j=0;j<7;j++) {
      if (fontData[j+4] != fontDataRom[j]) {
        verifyError ++ ;
        Serial.print("★") ;
      }
      sprintf(str,"%02X,",fontDataRom[j]) ;
      Serial.print(str) ;
    }
    if (verifyError > 0) {
      Serial.println(" -- ERROR") ;
    } else {
      Serial.println() ;
    }

    indexAddr += 2 ;
    fontAddr += 7 ;
  }
  Serial.print("--- verify --- ") ;
  if (verifyError > 0) {
    Serial.println("ERROR") ;
  } else {
    Serial.println("OK") ;
  }
}

// -------------------------------------------
// ROM 書き込み :繰り返しはしないので setup　のみ
// -------------------------------------------
void setup() {
  // initialize serial communications at 9600 bps:
  Serial.begin(9600);
  Wire.begin( ) ;
  
  Serial.println("\n---　漢字ROM Make ---") ;
  writeIndexTable(true) ;   // FONT INDEX テーブルの存在フラグ：初回のみ true、2回目以降false 
  writeHFTable(true) ;      // 半角カナ全角変換テーブル ：初回のみ true、2回目以降false

  // ----- FONTデータ
  // EEPROMに書き込むデータは全体で約90KByteあります。
  // Arduinoのプログラム領域は32KByteしかないですから、分割しないと入りません。
  // なので FONT00～FONTFFまでの10ブロックに分割してあります
  // MISAKI.h の　#defineのコメントを順番に外して計10回実行してください。
  
#ifdef FONT00
  writeFont(font00,sizeof(font00) / 11 , true) ;
#endif

#ifdef FONT20
  writeFont(font20,sizeof(font20) / 11 , true) ;
#endif  

#ifdef FONT30
  writeFont(font30,sizeof(font30) / 11 , true) ;
#endif  

#ifdef FONT40
  writeFont(font40,sizeof(font40) / 11 , true) ;
#endif  

#ifdef FONT50
  writeFont(font50,sizeof(font50) / 11 , true) ;
#endif

#ifdef FONT60
  writeFont(font60,sizeof(font60) / 11 , true) ;
#endif

#ifdef FONT70
  writeFont(font70,sizeof(font70) / 11 , true) ;
#endif

#ifdef FONT80
  writeFont(font80,sizeof(font80) / 11 , true) ;
#endif

#ifdef FONT90
  writeFont(font90,sizeof(font90) / 11 , true) ;
#endif

#ifdef FONTFF
  writeFont(fontFF,sizeof(fontFF) / 11 , true) ;
#endif
}

void loop() {
}
